# encoding: UTF-8
# typed: false

# Persistent-💎: Ruby gem for easily creating immutable data structures
# Copyright (c) 2017-2018 Ivo Anjo <ivo.anjo@ist.utl.pt>
#
# This file is part of Persistent-💎, and parts were inspired by Ruby's set.rb
# https://github.com/ruby/ruby/blob/v2_6_0_preview2/lib/set.rb
# Copyright (c) 2002-2016 Akinori MUSHA <knu@iDaemons.org>
# All rights reserved.  You can redistribute and/or modify it under the same terms as Ruby. See
# https://github.com/ruby/ruby/blob/v2_6_0_preview2/COPYING for more details.

# frozen_string_literal: true

require 'set'

module Persistent💎
  # Workaround for https://github.com/jruby/jruby/issues/5227
  # We monkey patch the Ruby Set methods affected to restore their interoperable behavior
  #
  module JRuby92SetWorkaround
    def self.workaround_needed?
      workaround_checker = Object.new
      workaround_checker.instance_eval do
        def is_a?(klass)
          super || klass == Set
        end

        def size
          0
        end

        def all?
          false
        end
      end

      begin
        Set.new.superset?(workaround_checker)
        false
      rescue ArgumentError => e
        e.message == "value must be a set"
      end
    end

    if RUBY_PLATFORM == 'java' && JRUBY_VERSION.start_with?('9.2.') && workaround_needed?
      class ::Set
        # Save existing Set methods

        alias persistent_dmnd_workaround_original_equal ==
        alias persistent_dmnd_workaround_original_superset? superset?
        alias persistent_dmnd_workaround_original_proper_superset? proper_superset?
        alias persistent_dmnd_workaround_original_subset? subset?
        alias persistent_dmnd_workaround_original_proper_subset? proper_subset?
        alias persistent_dmnd_workaround_original_intersect? intersect?
        alias persistent_dmnd_workaround_original_disjoint? disjoint?
        alias persistent_dmnd_workaround_original_flatten_merge flatten_merge
        alias persistent_dmnd_workaround_original_flatten flatten
        alias persistent_dmnd_workaround_original_flatten! flatten!

        # Redefine all set methods that use instanceof RubySet to restore previous behavior

        def ==(other)
          persistent_dmnd_workaround_original_equal(other) ||
            (other.is_a?(::Set) && size == other.size && other.all? { |o| include?(o) })
        end

        def superset?(set)
          persistent_dmnd_workaround_original_superset?(set)
        rescue ArgumentError => e
          raise unless e.message == "value must be a set" && set.is_a?(::Set)
          size >= set.size && set.all? { |o| include?(o) }
        end

        def proper_superset?(set)
          persistent_dmnd_workaround_original_proper_superset?(set)
        rescue ArgumentError => e
          raise unless e.message == "value must be a set" && set.is_a?(::Set)
          size > set.size && set.all? { |o| include?(o) }
        end

        def subset?(set)
          persistent_dmnd_workaround_original_subset?(set)
        rescue ArgumentError => e
          raise unless e.message == "value must be a set" && set.is_a?(::Set)
          size <= set.size && all? { |o| set.include?(o) }
        end

        def proper_subset?(set)
          persistent_dmnd_workaround_original_proper_subset?(set)
        rescue ArgumentError => e
          raise unless e.message == "value must be a set" && set.is_a?(::Set)
          size < set.size && all? { |o| set.include?(o) }
        end

        def intersect?(set)
          persistent_dmnd_workaround_original_intersect?(set)
        rescue ArgumentError => e
          raise unless e.message == "value must be a set" && set.is_a?(::Set)
          if size < set.size
            any? { |o| set.include?(o) }
          else
            set.any? { |o| include?(o) }
          end
        end

        def disjoint?(set)
          persistent_dmnd_workaround_original_disjoint?(set)
        rescue ArgumentError => e
          raise unless e.message == "value must be a set" && set.is_a?(::Set)
          !(
            if size < set.size
              any? { |o| set.include?(o) }
            else
              set.any? { |o| include?(o) }
            end
          )
        end

        def flatten_merge(set, seen = ::Set.new)
          set.each do |o|
            if o.is_a?(::Set)
              if seen.include?(o_id = o.object_id)
                raise ArgumentError, "tried to flatten recursive Set"
              end

              seen.add(o_id)
              flatten_merge(o, seen)
              seen.delete(o_id)
            else
              add(o)
            end
          end

          self
        end

        def flatten
          self.class.new.flatten_merge(self)
        end

        def flatten!
          replace(flatten()) if any? { |o| o.is_a?(::Set) }
        end

        # Reset aliases to point to redefined methods (otherwise they would keep on pointing to the old variants)

        alias >= superset?
        alias > proper_superset?
        alias <= subset?
        alias < proper_subset?
      end
    end
  end
end
