# encoding: UTF-8
# typed: false

# Persistent-💎: Ruby gem for easily creating immutable data structures
# Copyright (c) 2017 Ivo Anjo <ivo.anjo@ist.utl.pt>
#
# This file is part of Persistent-💎.
#
# MIT License
#
# Copyright (c) 2017 Ivo Anjo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# frozen_string_literal: true

module Persistent💎
  module JRubyWorkaround
    if RUBY_PLATFORM == 'java' && JRUBY_VERSION < "9.2.0.0" # Only supposed to do something on legacy JRuby versions
      BROKEN_ENCODING = "\x00"
      private_constant :BROKEN_ENCODING

      BROKEN_ENCODING_JRUBY_1_7 = [239, 146, 142].pack('c*').force_encoding('UTF-8')
      private_constant :BROKEN_ENCODING_JRUBY_1_7

      def self.included(klass)
        # Make methods on this module also available inside classes (e.g. not just in instances)
        klass.extend(self)
      end

      def method_added(name)
        super
        maybe_fix_broken_encoding(name)
      end

      def singleton_method_added(name)
        super
        maybe_fix_broken_encoding(name, self.singleton_class)
      end

      private

      # Workaround JRuby encoding bug
      # See https://github.com/jruby/jruby/issues/4878
      # and https://gitlab.com/ivoanjo/persistent-dmnd/issues/5
      #
      # The trick here is that on JRuby, methods with emojis in the name are defined, but the resulting name is mangled
      # so instead of for instance #to_💎 the result is #to_?\x00. (Or #💎ify will become #?ify\x00.) This behavior is
      # consistent between def and calling a method, so def will register a method #to_?\x00, and a normal method call
      # will do the same mangling and thus look for a method **WITH** the mangled name.
      #
      # But there are still other places in the code that rely on the correct name, such as alias and public_send/send,
      # so we additionally use this module to define the correct name as an alias for the broken name.
      #
      # Thus, it all works on JRuby, albeit in a very roundabout way!
      #
      # FOR JRUBY 1.7:
      # On JRuby 1.7 it seems that the mangling is different, and the emoji just gets replaced with ?, so for
      # instance #to_💎 becomes #to_?. But then, to make things interesting, on symbols created without quotes, it
      # gets mangled with the BROKEN_ENCODING_JRUBY_1_7, e.g :to_💎 becomes :to_BROKEN_ENCODING_JRUBY_1_7 (but
      # :"to_💎" is still correct).
      # To support both of these, we add an extra alias on JRuby, so that everything else seems to work even if it is
      # actually using the name with the mangled encoding.
      #
      def maybe_fix_broken_encoding(name, target = nil)
        name_string = name.to_s

        if JRUBY_VERSION.start_with?('1.7.')
          unless name_string.match(/\?.+/) ||
                 ['a?', 'h?', 's?', 'each?'].include?(name_string) ||
                 ['_?', '_a?', '_h?', '_s?'].any? { |suffix| name_string.end_with?(suffix) }
            return
          end
        else
          return unless name_string.end_with?(BROKEN_ENCODING)
        end

        fixed_name = name_string.sub('?', '💎').delete(BROKEN_ENCODING).to_sym
        fixed_name_jruby_1_7 = name_string.sub('?', BROKEN_ENCODING_JRUBY_1_7).to_sym

        if target
          target.instance_eval { alias_method(fixed_name, name) }
          target.instance_eval { alias_method(fixed_name_jruby_1_7, name) } if JRUBY_VERSION.start_with?('1.7.')
        else
          alias_method(fixed_name, name)
          alias_method(fixed_name_jruby_1_7, name) if JRUBY_VERSION.start_with?('1.7.')
        end
      end
    end
  end
end
